#include <stdio.h>
#include <stdlib.h>
#include <system.h>
#include <libfdt.h>
#include <smp.h>		/* cpu_online_map */

/* Some defines for FDT
 */
#define MAX_PROP_LEN     		256
#define CMDLINE_SIZE           	2048

/* Macros for accessing FDT
 */
enum {
        PROP_STR,
        PROP_INT32
};
#define GET_FPROP(path, prop, prop_buf)                                          \
	if (get_fdt_prop(path, prop, prop_buf, PROP_INT32) != 0) {	         \
		printf("Get prop (INT32) failed on %s (prop=%s)\n", path, prop); \
		return -1;                                                       \
	}

#define GET_FPROP_STR(path, prop, prop_buf)                                      \
        if (get_fdt_prop(path, prop, prop_buf, PROP_STR) != 0) {                 \
                printf("Get prop (STR) failed on %s (prop=%s)\n", path, prop);   \
                return -1;                                                       \
        }
/* FDT-related Global variables
 */
void *fdt;
char arcs_cmdline[CMDLINE_SIZE];
unsigned long fw_arg0, fw_arg1, fw_arg2, fw_arg3;
char app_mode[MAX_PROP_LEN];
char app_loadaddr[MAX_PROP_LEN];
int dom_destroy_hack = 0;

extern uint32_t cpu_online_map;
extern uint64_t heap_size;
extern uint64_t contig_alloc_off, contig_alloc_sz;
extern void add_mem_map(uint64_t addr, uint64_t size);
extern void uart_update_params(uint64_t sharedcfg, int id);

static __inline__ unsigned long long get_fdt_dword(unsigned int *prop)
{
       unsigned int *buf = prop;
#ifdef __MIPSEL__
    return ((uint64_t)buf[1] << 32 | buf[0]);
#else
        return ((uint64_t)buf[0] << 32 | buf[1]);
#endif
}

static int get_fdt_prop(const char* node_path,
						const char *prop, void *prop_buf, int type)
{
	int prop_len;
	const void *pbuf;

#ifdef __MIPSEL__ /* Little Endian */
	char * src, *dst;
	int i;
#endif

	int node = fdt_path_offset(fdt, node_path);

	if (node < 0) {
		return -1;
	}

	pbuf = (char *)fdt_getprop(fdt, node, prop, &prop_len);
	if (pbuf == NULL) {
		return -1;
	}
#ifdef __MIPSEL__ /* Little Endian */

	if (type == PROP_STR) {
		memcpy((void *) prop_buf, (void * )pbuf, prop_len);

	} else { /* Copy in reverse to account for host LE order */
		src = (char *) pbuf;
		dst = (char *) prop_buf;
		for (i = 0; i < prop_len; ++i) {
			dst[i] = src[prop_len - i - 1];
		}
	}

#else
	memcpy((void *) prop_buf, (void * )pbuf, prop_len);
#endif /* #ifdef (__MIPSEL__) */
	return 0;
}

int process_fdt(void)
{
	char dom_node_str[32];

	uint32_t na, ns, regs[4], entries = 1; /* move to FDT */
	int uart_id, fdt_err, i;
	uint32_t uart_sharedcfg;

	printf("(NetOS): fw_arg0=%lx, fw_arg1=%lx, fw_arg2=%lx\n", fw_arg0, fw_arg1, fw_arg2);

	fdt = (void *)fw_arg0;
	if (fdt == NULL) {
		return -1;
	}

	/* Convert the Application addr
	 * to KSEG0 physical
	 */
	fw_arg1 &= 0x1fffffffUL;

	dom_id = (int)fw_arg2;

	/* Validate FDT
	 */
	fdt_err = fdt_check_header(fdt);
	if (fdt_err < 0) {
		printf("FDT_ERROR: Invalid FDT Detected: %s\n",
				fdt_strerror(fdt_err));
		return -1;
	}

	snprintf(dom_node_str, 32, "/doms/dom@%d/uart", dom_id);
	GET_FPROP(dom_node_str,"id", &uart_id);

        if (get_fdt_prop(dom_node_str, "sharedcfg", &uart_sharedcfg, PROP_INT32) != 0) {
		/* Shared access to Uart */
		uart_sharedcfg = 0;
        }
	/* uart_sharedcfg == 0 ? Shared access : Exclusive access */
	uart_update_params((uint64_t)uart_sharedcfg, dom_id);

	snprintf(dom_node_str, 32, "/doms/dom@%d", dom_id);

	/* get the heap size */
	GET_FPROP(dom_node_str, "heap-size", regs);
	heap_size = get_fdt_dword(regs);
	GET_FPROP(dom_node_str, "shared-mem", regs);

#ifndef __MIPSEL__
	contig_alloc_off = (uint64_t) regs[0] << 32 | regs[1];
	contig_alloc_sz = (uint64_t) regs[2] << 32 | regs[3];
#else
	contig_alloc_off = get_fdt_dword(&regs[2]);
	contig_alloc_sz = get_fdt_dword(regs);
#endif
	/*printf("heap size %llx shared mem Addr %llx, Size %llx\n",
	  heap_size, contig_alloc_off, contig_alloc_sz);*/

	app_mode[0] = app_loadaddr[0] = 0;
	GET_FPROP_STR(dom_node_str, "app-bootcmd", app_mode);
	GET_FPROP_STR(dom_node_str, "app-loadaddr", app_loadaddr);

	/* extract CPU online info and UART
	 */
	GET_FPROP_STR("/chosen", "bootargs", arcs_cmdline);
	GET_FPROP_STR(dom_node_str, "device_type", arcs_cmdline);

	snprintf(dom_node_str, 32, "/doms/dom@%d/cpu", dom_id);
	GET_FPROP(dom_node_str, "onlinemask", &cpu_online_map);

	/* extract memory ranges,
	 */
	GET_FPROP("/doms", "#address-cells", &na);
	GET_FPROP("/doms", "#size-cells", &ns);

	snprintf(dom_node_str, 32, "/doms/dom@%d/memory", dom_id);
	GET_FPROP(dom_node_str, "reg", regs);

	for (i = 0; i < entries; i++)
		add_mem_map((uint64_t)regs[i*2], (uint64_t)regs[i*2+1]);

	snprintf(dom_node_str, 32, "/doms/dom@%d/kernel", dom_id);
	GET_FPROP(dom_node_str, "dom_destroy_hack", &dom_destroy_hack);

	return 0;
}
