/*
  FUNCTION
  <<__malloc_lock>>, <<__malloc_unlock>>---lock malloc pool

  INDEX
  __malloc_lock
  INDEX
  __malloc_unlock

  ANSI_SYNOPSIS
  #include <malloc.h>
  void __malloc_lock (struct _reent *<[reent]>);
  void __malloc_unlock (struct _reent *<[reent]>);

  TRAD_SYNOPSIS
  void __malloc_lock(<[reent]>)
  struct _reent *<[reent]>;

  void __malloc_unlock(<[reent]>)
  struct _reent *<[reent]>;

  DESCRIPTION
  The <<malloc>> family of routines call these functions when they need to lock
  the memory pool.  The version of these routines supplied in the library use
  the lock API defined in sys/lock.h.  If multiple threads of execution can
  call <<malloc>>, or if <<malloc>> can be called reentrantly, then you need to
  define your own versions of these functions in order to safely lock the
  memory pool during a call.  If you do not, the memory pool may become
  corrupted.

  A call to <<malloc>> may call <<__malloc_lock>> recursively; that is,
  the sequence of calls may go <<__malloc_lock>>, <<__malloc_lock>>,
  <<__malloc_unlock>>, <<__malloc_unlock>>.  Any implementation of these
  routines must be careful to avoid causing a thread to wait for a lock
  that it already holds.
*/

#include <malloc.h>
#include <sys/lock.h>
#include <sys/types.h>
#include <unistd.h>

struct malloc_lock
{
	volatile unsigned int id;
	volatile unsigned int count;
};

static struct malloc_lock mlock;

static inline void __spin_lock_r(int id)
{
	unsigned long long tmp;

	__asm__ volatile
		(
			".set push         \n"
			".set noreorder    \n"
			".set mips64       \n"
			"1: ll %1, %2      \n"
			"   beq %1, %3, 2f \n"
			"   sync           \n"
			"   bnez %1, 1b    \n"
			"   nop            \n"
			"   move %1, %3    \n"
			"   sc %1, %0      \n"
			"   beqz %1, 1b    \n"
			"   nop            \n"
			"2: nop            \n"
			".set pop          \n"
			:"=m"(mlock.id), "=&r"(tmp)
			:"m"(mlock.id), "r"(id)
			:"memory"
			);
	mlock.count++;
}

static inline void __spin_unlock_r(void)
{
	unsigned long long tmp;

	mlock.count--;

	if ((int)mlock.count < 0) {
		for(;;) { __asm__ __volatile__ ("wait\n"); }
	}

	__asm__ volatile
		(
			".set push 		\n"
			".set noreorder    	\n"
			".set mips64       	\n"
			"1: lw %1, %2      	\n"
			"   bnez %1, 2f         \n"
			"   sync                \n"
			"   sw $0, %0      	\n"
			"2: nop                 \n"
			".set pop               \n"
			:"=m"(mlock.id), "=&r"(tmp)
			:"m"(mlock.count)
			:"memory"
			);
}

void __malloc_lock (struct _reent *ptr)
{
	__spin_lock_r((getpid() + 1)); /*varies from 1 to 32 */
}

void __malloc_unlock (struct _reent *ptr)
{
	__spin_unlock_r();
}
