/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/
#include <stdio.h>
#include <stdlib.h>
#include <iomap.h>
#include "pic.h"
#include "irq.h"
#include "mm.h"
#include <stddef.h>
#include <mips-exts.h>
#include "spinlock.h"
#include "xlr_boot_lib.h"
#include "irqparam.h"
#include "cpu.h"

static spinlock_t irq_lock = SPIN_LOCK_UNLOCKED;
static int init_irq_done = 0;
extern irq_ops_t pic_ops;

static int startup_noop(int irq) { return 0; }
static void shutdown_noop(int irq) { }
static void epiack_noop(int irq) { }
static void proack_noop(int irq) { }
static void set_affinity_noop(int irq, unsigned int mask) { }
static irqreturn_t handler_noop(int irq, struct pt_regs *regs) { return 0; }
extern irqreturn_t netos_irq_handler(int irq, struct pt_regs *regs);

irq_ops_t irq_ops_noop = {
	.typename = "none",
	.startup = startup_noop,
	.shutdown = shutdown_noop,
	.epiack = epiack_noop,
	.proack = proack_noop,
	.set_affinity = set_affinity_noop,
	.handler = handler_noop,
};

irq_desc_t irqtable[MAX_CPUS][NR_IRQS];

// -------------------
//     IRQ Handler
// -------------------
void do_phnx_irq(int irq, struct pt_regs *regs)
{
	irq_desc_t *desc = &(irqtable[processor_id()][irq]);
	irqreturn_t retval;

	if (desc->status & IRQ_DISABLED) {
		printf("BooooG @ %s(%s:%d)\n", __FUNCTION__, __FILE__, __LINE__);
	}

	/* Step 1: Ack the pic */
	if (desc->ops.proack)
		desc->ops.proack(irq);

	if (desc->ops.handler)
		retval = desc->ops.handler(irq, regs);

	/* Step 3: Ack the pic */
	if (desc->ops.epiack)
		desc->ops.epiack(irq);
}

// -----------
//     API
// -----------

int request_irq(unsigned int irq,
		irqreturn_t (*handler)(int, struct pt_regs *),
		unsigned long irqflags,
		unsigned int mask)
{
	irq_desc_t *desc = &irqtable[processor_id()][irq];
	unsigned int flags;
	int retval = 0;

	if (irq >= NR_IRQS) {
		printf ("%s(): irq exceeds highest irq number (%d)\n", __FUNCTION__, NR_IRQS);
		return -1;
	}

	if (mask == 0)
		mask = (1 << processor_id());

	spin_lock_irqsave(&irq_lock, flags);
	write_phnx_eimr(read_phnx_eimr() | (1ULL << irq)); // +++
	spin_unlock_irqrestore(&irq_lock, flags);

	spin_lock_irqsave(&(desc->lock), flags);

	/* if applications sharing the same data / code space request for 
	   the same IRQ, don't treat it as error
	 */
	if(desc->ops.handler == handler) {
		spin_unlock_irqrestore(&(desc->lock), flags);
		return 0;
	}
	if (desc->ops.handler != NULL) {
		if((desc->flags & IRQ_FLAGS_RESERVED) == IRQ_FLAGS_RESERVED) {
			printf("Error: Cannot request for a Reserved IRQ\n");
			spin_unlock_irqrestore(&(desc->lock), flags);
			return -1;
		}
	/*	if(desc->ops.handler != handler_noop)
			printf("Overwriting existing handler\n"); */
	}

	desc->ops.handler = handler;
	desc->flags = irqflags;
	desc->mask = mask;
	desc->status &= ~IRQ_DISABLED;

	if (desc->ops.startup) {
		retval = desc->ops.startup(irq);
		if (retval)
			goto ret;
	}
	if (desc->ops.set_affinity)
		desc->ops.set_affinity(irq, mask);

ret: 
	spin_unlock_irqrestore(&(desc->lock), flags);

	return retval;
}

void free_irq(unsigned int irq)
{
	irq_desc_t *desc;
	unsigned int flags;

	if (irq >= NR_IRQS) {
		printf ("%s(): irq exceeds highest irq number (%d)\n", __FUNCTION__, NR_IRQS);
		return;
	}

	desc = &irqtable[processor_id()][irq];

	/* we grab the big lock as we modify eimr too */

	spin_lock_irqsave(&irq_lock, flags);
	write_phnx_eimr(read_phnx_eimr() & (~(1ULL << irq)));
	spin_unlock_irqrestore(&irq_lock, flags);

	spin_lock_irqsave(&(desc->lock), flags);

	desc->status |= IRQ_DISABLED;
	if (desc->ops.shutdown)
		desc->ops.shutdown(irq);
	desc->ops = irq_ops_noop;
	desc->flags = desc->mask = 0;

	spin_unlock_irqrestore(&(desc->lock), flags);	

	return;
}

void set_irqops(int irq, int option, void *function)
{
	unsigned int flags;
	int cpu = processor_id();

	/* check for irq number validity */
	if (irq >= NR_IRQS) {
		printf ("%s(): irq exceeds highest irq number (%d)\n", __FUNCTION__, NR_IRQS);
		return;
	}

	/* lock irqtable */	
	spin_lock_irqsave(&(irqtable[cpu][irq].lock), flags);

	switch (option)
	{
		case NLM_SET_STARTUP:
			irqtable[cpu][irq].ops.startup = (int (*)(int)) function;
			break;

		case NLM_SET_SHUTDOWN:
			irqtable[cpu][irq].ops.shutdown = (void (*)(int)) function;
			break;

		case NLM_SET_EPIACK:
			irqtable[cpu][irq].ops.epiack = (void (*)(int)) function;
			break;

		case NLM_SET_PROACK:
			irqtable[cpu][irq].ops.proack = (void (*)(int)) function;
			break;

		case NLM_SET_AFFINITY:
			irqtable[cpu][irq].ops.set_affinity = (void (*)(int, unsigned int)) function;
			break;
	}

	/* unlock irqtable */
	spin_unlock_irqrestore(&(irqtable[cpu][irq].lock), flags);
}


/*
 * INITIALIZATION ROUTINES. NEED TO BE CALLED FROM HARDWARE_INIT_HOOK
 */


int init_irq(void)
{
	/* Initialize the irq descriptors */
	int i, cpu;
	int rv;

	spin_lock(&irq_lock);

	if (!init_irq_done) {
		for(cpu = 0; cpu < MAX_CPUS; cpu++) {
			for (i = 0; i < NR_IRQS; i++)
				irqtable[cpu][i].ops = irq_ops_noop;

			for (i = PIC_IRT_FIRST_IRQ; i <= PIC_IRT_LAST_IRQ; ++i)
				irqtable[cpu][i].ops = pic_ops;
		}
		init_irq_done = 1;
	}
	spin_unlock(&irq_lock);

	/*Intialize all the netos handled intr */
	cpu = processor_id();
	rv = request_irq(IRQ_UART, netos_irq_handler, 0, 1 << cpu);
	if(rv)
		goto err_exit;
	rv = request_irq(IRQ_WAKEUP_CPU_IPI, netos_irq_handler, 0, 1 << cpu);
	if(rv)
		goto err_exit;
	rv = request_irq(IRQ_STOP_CPU_IPI, netos_irq_handler, 0, 1 << cpu);
	if(rv)
		goto err_exit;
	rv = request_irq(IRQ_RECLAIM_CPU_RSRC_IPI, netos_irq_handler, 0, 1 << cpu);
	if(rv)
		goto err_exit;
	rv = request_irq(IRQ_JOIN_CPU_IPI, netos_irq_handler, 0, 1 << cpu);
	if(rv)
		goto err_exit;
	rv = request_irq(IRQ_MAC, netos_irq_handler, 0, 1 << cpu);
	if(rv)
		goto err_exit;
	rv = request_irq(IRQ_DESTROY_DOMAIN_IPI, netos_irq_handler, 0, 1 << cpu);
	if(rv)
		goto err_exit;
	rv = request_irq(IRQ_TLB_FLUSH_IPI, netos_irq_handler, 0, 1 << cpu);
	if(rv)
		goto err_exit;


	return 0;
err_exit:
	printf("Error in irq init\n");
	cpu_halt();
	return -1;
}

int config_irq_sys_call(void *arg)
{
	struct irq_param *s = arg;
	int rv;

	switch(s->cmd)
	{
		case NLM_REQUEST_IRQ:
			rv = request_irq(s->irq, (void *)s->handler,  s->irqflags, s->mask);
			break;
		case NLM_FREE_IRQ:
			free_irq(s->irq);
			break;
		case NLM_SET_IRQOPS:
			set_irqops(s->irq, s->options, (void *)s->handler);
			break;
		default:
			s->rv = -1;
			return -1;
	}
	s->rv = 0;
	return 0;
}
