/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/

#include <stdio.h>
#include <stdlib.h>

#include <system.h>
#include <iomap.h>
#include "irq.h"
#include <stddef.h>
#include <mips-exts.h>
#include "spinlock.h"
#include "xlr_boot_lib.h"

#ifdef XLP_SIM
#include "nlm_hal_pic.h"
#else
#include "pic.h"
#endif

static spinlock_t pic_lock = SPIN_LOCK_UNLOCKED;

#if defined(_MIPS_ARCH_XLP)
/* Some routines Imported from elfwb 
 */
uint64_t read_timer_cycles(void)
{
	return TIMER_CYCLES_MAXVAL - nlm_hal_pic_read_timer(6);
}

double read_timer_period(void)
{   
	const double period = 7.5 * 1.0e-9; /* 133.33 MHz */
	return period;
}   

double read_timer_seconds(void)
{
	uint64_t cycles = read_timer_cycles();
	const double period = read_timer_period();
	return (double)cycles * period;
}

void setup_timer(int cpu)
{
#if 0
	const int timer = 6;

	/* Use PIC Timer 6 as a free running counter */
	nlm_hal_pic_write_timer(timer, TIMER_CYCLES_MAXVAL);

	/* we Don't need interrupts */
	nlm_hal_pic_write_irt_direct(PIC_IRT_TIMER_6, 
			IRT_VALID, 0, LOCAL_SCHEDULING, 
			PIC_TIMER_6_IRQ,
			processor_id());
	
	/* Enable the timer */
	nlm_hal_pic_update_control(1 << (10 + timer));
#endif
}
#else
uint64_t read_timer_cycles(void)
{
        return 0; /* TODO */
}

double read_timer_period(void)
{   
        return 0; /* TODO */
}   

double read_timer_seconds(void)
{
        return 0; /* TODO */
}

void setup_timer(int cpu)
{
}
#endif /* _MIPS_ARCH_XLP */


/* PIC IRQ Handler/Init/Ack Routines
 */
int pic_startup(int irq)
{
    phoenix_reg_t *mmio = phoenix_io_mmio(PHOENIX_IO_PIC_OFFSET);
    unsigned int flags;
    
    int level = 0;
    
    if (!PIC_IRQ_IS_IRT(irq)) {
        printf ("%s: Requested irq does not correspond to an IRT\n", __FUNCTION__);
        return 1;
    }

    spin_lock_irqsave(&pic_lock, flags);

#ifndef XLP_SIM
    if(!PIC_IRQ_IS_EDGE_TRIGGERED(irq))
        level = 1;
#endif
	/* 
	 * By default all the interrupts are initialized as 
	 * level-senstive ; fix for the PCMCIA flash
	 */

    /* i2c needs the PIC to be configured as EDGE-triggered, otherwise PIC gets continuous interrupts */
    if(irq == IRQ_PIC_I2C_0 || irq == IRQ_PIC_I2C_1)
        level = 0;

#ifdef XLP_SIM
    mmio = 0; /* For compiler sake */
    nlm_hal_pic_write_irt_direct(irq - PIC_IRQ_BASE, IRT_VALID, 0, LOCAL_SCHEDULING, irq, 0);

#else
    phoenix_write_reg(mmio, PIC_IRT_1_BASE + irq - PIC_IRQ_BASE,
					  (IRT_VALID << 31) | (level << 30) |
					  (LOCAL_SCHEDULING << 6)| irq );
#endif
    spin_unlock_irqrestore(&pic_lock, flags);
	
    return 0;
}

void pic_ack(int irq)
{
    phoenix_reg_t    *mmio, *pci_mmio;
    int               i;
    phoenix_reg_t     reg;
    unsigned int      flags;

    mmio = phoenix_io_mmio(PHOENIX_IO_PIC_OFFSET);
    pci_mmio = phoenix_io_mmio(PHOENIX_IO_PCIX_OFFSET);
    if (!PIC_IRQ_IS_IRT(irq)) {
        printf ("%s: Requested irq does not correspond to an IRT\n", __FUNCTION__);
        return;
    }

    /* 
	 * Interrupt (level sensitive) acknowledge 
	 * method for the PCMCIA flash 
	 */
    if (irq == 21 ) {
        reg = *(unsigned char *)(0xBD0001f7);
        reg = *(unsigned long *)(0xBEF19180);
        for (i = 0; i < 0x100; i++)
            ;
        *(unsigned long *)(0xBEF19180) = reg ;
        for (i = 0; i < 0x1000; i++)
            ;
        reg = *(unsigned long *)(0xBEF19180);
    }

    /* Deal with All PCI-Interrupts.. Brigde ACK */
    if ((!is_xls()) && (irq == 24)) {
        phoenix_read_reg(pci_mmio, (0x140 >> 2));
    }

    /*
	 * If edge triggered IRQ, ack it immediately, 
	 * else when the device interrupt condition
	 * is cleared, we may lose interrupts
	 */
#ifndef XLP_SIM
    if (PIC_IRQ_IS_EDGE_TRIGGERED(irq)) {
        spin_lock_irqsave(&pic_lock, flags);
        phoenix_write_reg(mmio, PIC_INT_ACK, (1 << (irq - PIC_IRQ_BASE)));
        spin_unlock_irqrestore(&pic_lock, flags);
    }
#else
    flags = 0; /* To silence compiler */
#endif
}

void pic_end(int irq)
{
    phoenix_reg_t *mmio;
    unsigned int flags;

    mmio = phoenix_io_mmio(PHOENIX_IO_PIC_OFFSET);
    if (!PIC_IRQ_IS_IRT(irq)) {
        printf ("%s: Requested irq does not correspond to an IRT\n", __FUNCTION__);
        return;
    }

    /* 
	 * If level triggered, ack it after 
	 * the device condition is cleared 
	 */
    if (!PIC_IRQ_IS_EDGE_TRIGGERED(irq)) {
        spin_lock_irqsave(&pic_lock, flags);
#ifdef XLP_SIM
	nlm_hal_ack_pic(nlm_hal_irq_to_irt(irq));
#else
        phoenix_write_reg(mmio, PIC_INT_ACK, (1 << (irq - PIC_IRQ_BASE)));
#endif
        spin_unlock_irqrestore(&pic_lock, flags);
    }
}

static void pic_shutdown(int irq)
{
    phoenix_reg_t   *mmio;
    phoenix_reg_t    reg;
    unsigned int     flags;

    mmio = phoenix_io_mmio(PHOENIX_IO_PIC_OFFSET);
    if (!PIC_IRQ_IS_IRT(irq))
        return;

    spin_lock_irqsave(&pic_lock, flags);

    /* 
	 * What happens if this irq is currently pending 
	 * an ack ? Assume, that doesn't happen ?
	 */
#ifdef XLP_SIM
    reg = (__u32) nlm_hal_read_pic_reg((pic_reg_t *)PHOENIX_IO_PIC_OFFSET, irq - PIC_IRQ_BASE);
    nlm_hal_write_pic_reg((pic_reg_t *) PHOENIX_IO_PIC_OFFSET, irq - PIC_IRQ_BASE, (reg & ~(1<<31)));

#else
    reg = phoenix_read_reg(mmio, PIC_IRT_1_BASE + irq - PIC_IRQ_BASE);
    phoenix_write_reg(mmio, PIC_IRT_1_BASE + irq - PIC_IRQ_BASE, (reg & ~(1<<31)));
#endif

    spin_unlock_irqrestore(&pic_lock, flags);
}

void pic_set_affinity(int irq, unsigned int mask)
{
    phoenix_reg_t *mmio;
    unsigned int flags;

    mmio = phoenix_io_mmio(PHOENIX_IO_PIC_OFFSET);
    if (!PIC_IRQ_IS_IRT(irq))
        return;

    spin_lock_irqsave(&pic_lock, flags);
#ifdef XLP_SIM
    nlm_hal_write_pic_reg((pic_reg_t *) XLP_IO_PIC_OFFSET, irq - PIC_IRQ_BASE, mask);
#else
    phoenix_write_reg(mmio, PIC_IRT_0_BASE + irq - PIC_IRQ_BASE, mask);
#endif
    spin_unlock_irqrestore(&pic_lock, flags);
}

irqreturn_t pic_handler(int irq, struct pt_regs *regs)
{
	return 0;
}

irq_ops_t pic_ops = {
	.typename = "NLM-Phoniex PIC",
    .startup = pic_startup,
    .shutdown = pic_shutdown,
    .epiack = pic_end,
    .proack = pic_ack,
    .set_affinity = pic_set_affinity,
	.handler = pic_handler
};
