#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "cpu.h"
#include "xlr_boot_lib.h"
#include "mm.h"
#include "irq.h"
#include "atomic.h"

static spinlock_t create_mapping_lock;
atomic_t flushtlb_cpucnt = { 0 };
unsigned long flushtlb_svaddr, flushtlb_evaddr;
extern uint32_t cpu_buddy_map;
void nlm_dump_tlb(void);

//#define DEBUG_TLB 1
/* Following var indicates that we have been launched from bootloader
 * which passed info about tlb list, so that we can use tlb
 * refill handler
 */
int tlb_refill_in_use = 0;
#define dbg_printf(fmt, args...)

extern void setup_tlb_uncached(unsigned long virt, unsigned long long phys, int size);
extern void show_regs(struct pt_regs *regs);
extern void handle_tlb_invalid();
extern void dump_cpu_info(int argc, char *argv[]);
extern void dump_tlb();

static unsigned long find_tlb_alignment(unsigned long long phys, 
					unsigned long size);
extern void copy_tlb_exception_handlers();
void setup_tlb_in_pte(unsigned long virt, unsigned long long phys,
		      unsigned int size, int attr);
extern struct per_cpu_info_loader per_cpu_info[];
extern volatile struct xlr_vcpu_wakeup_info *xlr_vcpu_wakeup_info;
extern void print_regs(struct pt_regs *regs, int cpu);
extern uint32_t cpu_online_map;
extern unsigned long tlb_align_size(uint64_t size);

#define nlm_align(x, y)     ((x) & (~((y)-1)))
#define nlm_roundup(x, y)   (nlm_align((x)+(y)-1, (y)))


void flush_tlb_range(unsigned long svaddr, unsigned long evaddr, uint32_t vpgsz)
{
	int idx;
	unsigned long vaddr;
	dbg_printf("%s in svaddr %lx evaddr %lx vpgsz %x\n", __FUNCTION__,
			svaddr, evaddr, vpgsz);
	//nlm_dump_tlb();
	
	vaddr = nlm_align(svaddr, vpgsz);
	while(vaddr < evaddr) {
		write_c0_entryhi(vaddr);
		tlb_probe();
		idx = read_c0_index();
		if(idx >= 0) {
			write_c0_entrylo0(0);
			write_c0_entrylo1(0);
			tlb_write_indexed();
		}
		dbg_printf("cpu %d  flush addr %lx\n", processor_id(), vaddr);
		vaddr += vpgsz;
	}
	dbg_printf("%s exit svaddr %lx evaddr %lx vpgsz %x\n", __FUNCTION__,
			svaddr, evaddr, vpgsz);
	//nlm_dump_tlb();
}

int find_and_set_wired_tlbs(long eaddr, struct per_cpu_info *my_info,
		struct xlr_vcpu_wakeup_info *t)
{
	struct ext_tlb *my_tlb;
	int i;
	struct cpu_tlb_mapping *map; 

	map = (struct cpu_tlb_mapping *)t->map;

#ifdef DEBUG_TLB
	printf("eaddr = %lx\n",eaddr);
#endif /* DEBUG_TLB */

	my_tlb = my_info->ext_tlb_ptr;


	for (i = 0; i < my_info->num_tlbs; i++) {
		if ((long)eaddr < my_tlb->virt + my_tlb->page_size * 2) {
			write_32bit_cp0_register(CP0_WIRED, i + 1);
			if (i + 1 > MAX_WIRED_ENTRIES) {
				printf("WARN:eaddr doesnt fall within "
				       "%d tlbs\n",MAX_WIRED_ENTRIES);
				return 0;
			}
			return i + 1;
		}
		my_tlb++;
	}

	printf("BUG: Didnt find eaddr within the tlbs\n");
	return 0;
}

void pte_init(int cpu, long eaddr)
{
	struct per_cpu_info *my_info = (void *)(per_cpu_info + cpu);
	struct ext_tlb *my_tlb;
	uint64_t attr = (3 << 3) | (1 << 2) | (1 << 1) | (1 << 0);
	int i;
	struct cpu_tlb_mapping *map;
	struct xlr_vcpu_wakeup_info *t;
	struct tlb_struct *pte;
	uint32_t ptesize;
	static int cpu_cnt = 0;
	unsigned long long pbase;

	ptesize = sizeof(struct tlb_struct) * PHYS_VIRT_NUM_PTE;

	t = (void *)(xlr_vcpu_wakeup_info  + cpu);
	map = (struct cpu_tlb_mapping *)t->map;

	tlb_refill_in_use = 1;
	find_and_set_wired_tlbs(eaddr, my_info, t);

	spin_lock(&create_mapping_lock);
	if(smp_mode) {
		pbase = ptebase;
		write_64bit_cp0_register_sel(CP0_OSSCRATCH, pbase, HPEXEC_PTEBASE_SCRATCH_IDX);
		cpu_cnt++;
		if(cpu_cnt > 1) {
			spin_unlock(&create_mapping_lock);
			return;
		}
	} else { 
		pbase = ptebase + ptesize * cpu_cnt;
		cpu_cnt++;
		write_64bit_cp0_register_sel(CP0_OSSCRATCH, pbase, HPEXEC_PTEBASE_SCRATCH_IDX);
	}
	spin_unlock(&create_mapping_lock);

	pte = (void *)(long)pbase;
	/* Setup all the tlb entries as invalid */
	for (i = 0; i < PHYS_VIRT_NUM_PTE; i++) {
		pte[i].entrylo0 = pte[i].entrylo1 = 0; 
	}

	my_tlb = my_info->ext_tlb_ptr;

	/* Fill in the entries... */
	for (i = 0; i < my_info->num_tlbs; i++) {

		setup_tlb_in_pte(my_tlb->virt, my_tlb->phys, my_tlb->page_size, attr);

#ifdef DEBUG_TLB
		printf("pte_init #%d: virt = %llx, phys = %llx, size = %llx\n",i,
		       (long long)my_tlb->virt,
		       (long long)my_tlb->phys,
		       (long long)my_tlb->page_size);
#endif /* DEBUG_TLB */
		my_tlb++;

	}
}
#ifndef XLP_SIM
unsigned long tlb_size_to_page_size(unsigned long size)
{
    if (size <= (4*1024)) return 4*1024;
    if (size <= (16*1024)) return 16*1024;
    if (size <= (64*1024)) return 64*1024;
    if (size <= (256*1024)) return 256*1024;
    if (size <= (1024*1024)) return 1024*1024;
    if (size <= (4*1024*1024)) return 4*1024*1024;
    if (size <= (16*1024*1024)) return 16*1024*1024;
    if (size <= (64*1024*1024)) return 64*1024*1024;

    return 256*1024*1024;
}

unsigned long tlb_size_to_mask(unsigned long size)
{
    if (size <= (4*1024)) return 0x0 << 13; 
    if (size <= (16*1024)) return 0x03 << 13; 
    if (size <= (64*1024)) return 0x0f << 13; 
    if (size <= (256*1024)) return 0x3f << 13; 
    if (size <= (1024*1024)) return 0xff << 13; 
    if (size <= (4*1024*1024)) return 0x3ff << 13; 
    if (size <= (16*1024*1024)) return 0xfff << 13; 
    if (size <= (64*1024*1024)) return 0x3fff << 13; 

    return 0xffff << 13; 
}
#endif

void setup_tlb_in_pte(unsigned long virt, unsigned long long phys, 
		      unsigned int size, int attr)
{
	int page_size;
	int page_mask;
	long i,k;
	struct tlb_struct *pte;

	if (!tlb_refill_in_use)
		return;

	pte = (void *)read_64bit_cp0_register_sel(CP0_OSSCRATCH,HPEXEC_PTEBASE_SCRATCH_IDX);

	page_size = tlb_size_to_page_size(size);
	page_mask = tlb_size_to_mask(page_size);

	for (i = virt >> (MIN_PAGE_BITS + 1),k=0; i < (virt + size * 2) >> 
	     (MIN_PAGE_BITS + 1); i++,k++) {
		dbg_printf("setup before idx %ld lo %08llx:%08llx physlo %08llx\n",
					i, pte[i].entrylo0,pte[i].entrylo1,pte[(phys >> 21) + k].entrylo1);
		
		pte[i].entrylo0 = phys >> 12 << 6 | attr;
		pte[i].entrylo1 |= (phys + page_size) >> 12 << 6 | attr;

		/* Zero out 34:63 of entry lo 0 and 1. */
		/* 
		 * Note: Entry lo1 bits 34:63 is overloaded. Dont zero them out
		 * 0:33->entrylo1
		 * 34:63->physical to virt mapping
		 */
		pte[i].entrylo0 = pte[i].entrylo0 << 30 >> 30 ;

		// entry_lo0 is stored as: 
		// 0:33->entrylo0 
		// 34:63->page_mask
		// See the tlb refill handler...
		pte[i].entrylo0 |= (unsigned long long)page_mask << 34;
		// Following is used for phys to virt translation
		pte[(phys >> 21) + k].entrylo1 |= (unsigned long long)i << 34;
		dbg_printf("setup after idx %ld lo %08llx:%08llx physlo %08llx\n",
					i, pte[i].entrylo0,pte[i].entrylo1,pte[(phys >> 21) + k].entrylo1);

	}
}


void dump_page_table()
{
	int i;
	struct tlb_struct *pte;
	pte = (void *)read_64bit_cp0_register_sel(CP0_OSSCRATCH,  HPEXEC_PTEBASE_SCRATCH_IDX);

	printf("Non-filled entries are not shown\n");
	printf("Size of each page entry = 0x%lx * 2\n\n",(long)1 << MIN_PAGE_BITS);
	printf("Index    Entrylo0    Entrylo1    Pagemask  Phys->Virt\n");
	for (i = 0; i < PHYS_VIRT_NUM_PTE; i++) {
		if (!(pte[i].entrylo0 || pte[i].entrylo1))
			continue;
		printf("%5d, %10llx, %10llx, %10llx, %10llx\n",i, pte[i].entrylo0 << 30 >> 30,
		       pte[i].entrylo1 << 30 >> 30, pte[i].entrylo0 >> 34, pte[i].entrylo1 >> 34 << 21);
	}

}


void tlb_invalid(struct pt_regs *regs)
{
	int cpu = processor_id();

	printf("(cpu%d) Got tlb invalid exception\n", cpu);

	print_regs(regs, cpu);

	enable_destroy_ipi();

	cpu_halt();
}



/* Return 0 on success */
/* Else return the index where virt address is already in use */
static int scan_pte_for_unmapped_region(int start, int count)
{
	int i;
	struct tlb_struct *pte;
	pte = (void *)read_64bit_cp0_register_sel(CP0_OSSCRATCH, HPEXEC_PTEBASE_SCRATCH_IDX);

	for (i = start; i < start + count; i++) {
		if (pte[i].entrylo0) 
			return i; // Failed... return failure index 
	}

	return 0;
}

/* Return unmapped virt addr or return 0 */
/* phys -> Physical addr
 * size -> How much memory needs to be mapped
 * Phys can be aligned or unaligned. Virt should be aligned in the same way.
 */
unsigned long find_unmapped_virt_addr(unsigned long long phys, 
					     unsigned long size)
{
	int i, x;
	int num_tlbs, initial_tlbs, max_tlbs;
	unsigned long mapped_size, this_size, max;

	num_tlbs = size >> (MIN_PAGE_BITS + 1);

	if (!num_tlbs)
		return 0;

	mapped_size = 0;
	max = 0;
	while (size > mapped_size) {
		this_size = find_tlb_alignment(phys + mapped_size, 
						    size - mapped_size);
		mapped_size += this_size;
		if (this_size > max)
			max = this_size;
	}

	max_tlbs = max >> (MIN_PAGE_BITS + 1);
	initial_tlbs = (phys & (max - 1)) >> (MIN_PAGE_BITS + 1);
	if (!initial_tlbs)
		initial_tlbs = max_tlbs;

	for (i=initial_tlbs; i <= NUM_PTE - num_tlbs; i+=max_tlbs) {
		x = scan_pte_for_unmapped_region(i, num_tlbs);
		/* if success... return here. */
		if (!x) return (unsigned long)i << (MIN_PAGE_BITS + 1);

		/* Failure case... x contains the failed entry */
		while (x >= i) /* Skip the failed entries */
			i += max_tlbs;

		/* Step back, because as part of the for loop end condition
		 * we are stepping again
		 */
		i -= max_tlbs;
	}
	/* Failed to find a contiguous virt space */
	return 0;
}

/* Given the phys address and size 
 * return the best possible first (tlb size * 2) for this combination
 * alignment assumed to be (2 * min_page_size); if not then return 0
 */
static unsigned long find_tlb_alignment(unsigned long long phys, 
					unsigned long size)
{
	int num_tlbs, i;
	
	num_tlbs = size >> (MIN_PAGE_BITS + 1);

	for (i = MIN_PAGE_BITS; i <= MAX_PAGE_BITS; i += 2) {
		if (phys & ((1 << (i + 1)) - 1))
			break;
		if (size < (1 << (i + 1)))
			break;
	}

	/* Case where phys or size is not aligned to (min_page_size * 2) */
	if (i == MIN_PAGE_BITS)
		return 0;

	/* i is 2 more than what it should be */
	/* So, shift is (i - 1) instead of (i + 1) */
	return (1 << (i - 1));
		
}

/* Returns virt addr corresponding to the created mapping */
/* If mapping is not created, return val is 0 */
/* size and phys should be aligned to (2*min_page_size) */
/* If this alignment is not followed, return val is 0 */
unsigned long create_mapping(unsigned long long phys, unsigned long size)
{
	int num_tlbs;
	long virt, mapped_size, this_size;

	/* return 0 if alignment is not proper */
	if (size % (1 << (MIN_PAGE_BITS + 1)))
		return 0;
	if (phys % (1 << (MIN_PAGE_BITS + 1)))
		return 0;

	num_tlbs = size >> (MIN_PAGE_BITS + 1);
	if (!num_tlbs)
		return 0;

	spin_lock(&create_mapping_lock);
	virt = find_unmapped_virt_addr(phys, size);
	if (!virt) {
		spin_unlock(&create_mapping_lock);
		return 0;
	}

	/* Generate optimum tlb sizes here */

	mapped_size = 0;
	while (size > mapped_size) {
		this_size = find_tlb_alignment(phys + mapped_size, 
						    size - mapped_size);
		setup_tlb(virt + mapped_size, phys + mapped_size, this_size/2);
#ifdef DEBUG_TLB
		printf("best mapping... phys %llx, virt %lx, size %lx\n",
		       phys + mapped_size,  virt + mapped_size, this_size);
#endif /* DEBUG_TLB */
		mapped_size += this_size;
	}
	spin_unlock(&create_mapping_lock);
	return virt;
}

/* Same as create_mapping... but creates uncached mapping */
unsigned long create_mapping_uncached(unsigned long long phys, unsigned long size)
{
	int num_tlbs;
	long virt, mapped_size, this_size;

	/* return 0 if alignment is not proper */
	if (size % (1 << (MIN_PAGE_BITS + 1)))
		return 0;
	if (phys % (1 << (MIN_PAGE_BITS + 1)))
		return 0;

	num_tlbs = size >> (MIN_PAGE_BITS + 1);
	if (!num_tlbs)
		return 0;

	spin_lock(&create_mapping_lock);
	virt = find_unmapped_virt_addr(phys, size);
	if (!virt) {
		spin_unlock(&create_mapping_lock);
		return 0;
	}

	/* Generate optimum tlb sizes here */

	mapped_size = 0;
	while (size > mapped_size) {
		this_size = find_tlb_alignment(phys + mapped_size, 
						    size - mapped_size);
		setup_tlb_uncached(virt + mapped_size, phys + mapped_size, this_size/2);
#ifdef DEBUG_TLB
		printf("best mapping... phys %llx, virt %lx, size %lx\n",
		       phys + mapped_size,  virt + mapped_size, this_size);
#endif /* DEBUG_TLB */
		mapped_size += this_size;
	}
	spin_unlock(&create_mapping_lock);
	return virt;
}

void setup_tlb(unsigned long virt, unsigned long long phys, int size)
{
	uint64_t attr = (3 << 3) | (1 << 2) | (1 << 1) | (1 << 0);
	return setup_tlb_in_pte(virt, phys, size, attr);
}

void setup_tlb_uncached(unsigned long virt, unsigned long long phys, int size)
{
	uint64_t attr = (2<<3)|(1<<2)|(1<<1)|(1<<0);
	return setup_tlb_in_pte(virt, phys, size, attr);
}

void hpexec_setup_tlb(unsigned long virt, uint64_t phys, uint64_t size)
{
	uint64_t mapped_size, this_size;

	mapped_size = 0;
	while (size > mapped_size) {
		this_size = find_tlb_alignment(phys + mapped_size, 
						    size - mapped_size);
		dbg_printf("hpexec_setup_tlb... phys %llx, virt %lx, size %llx\n",
		       (unsigned long long)(phys + mapped_size),  virt + mapped_size, 
		       (unsigned long long)this_size);

		setup_tlb(virt + mapped_size, phys + mapped_size, this_size/2);
		mapped_size += this_size;
	}
}

void clear_tlb_in_pte(unsigned long virt, unsigned long long phys, unsigned int size)
{
	int page_size;
	int page_mask;
	long i,k;
	struct tlb_struct *pte;
	uint64_t einfo_mask =  ((1ULL << 34) - 1);
	if (!tlb_refill_in_use)
		return;

	pte = (void *)read_64bit_cp0_register_sel(CP0_OSSCRATCH,HPEXEC_PTEBASE_SCRATCH_IDX);

	page_size = tlb_size_to_page_size(size);
	page_mask = tlb_size_to_mask(page_size);

	for (i = virt >> (MIN_PAGE_BITS + 1),k=0; i < (virt + size * 2) >> 
	     (MIN_PAGE_BITS + 1); i++,k++) {
		dbg_printf("clear before idx %ld lo %08llx:%08llx physlo %08llx\n",
				i, pte[i].entrylo0,pte[i].entrylo1,pte[(phys >> 21) + k].entrylo1);

		/* 
		 * Note: Entry lo1 bits 34:63 is overloaded. Dont zero them out
		 * 0:33->entrylo1
		 * 34:63->physical to virt mapping
		 */
		pte[i].entrylo1 &= (~einfo_mask); /* keep the upper bits as it is */
		pte[(phys >> 21) + k].entrylo1 &= einfo_mask; /* keep the lower bits as it is */

		// entry_lo0 is stored as: 
		// 0:33->entrylo0 
		// 34:63->page_mask
		// See the tlb refill handler...
		pte[i].entrylo0 = 0;
		dbg_printf("clear after idx %ld lo %08llx:%08llx physlo %08llx\n",
				i, pte[i].entrylo0,pte[i].entrylo1,pte[(phys >> 21) + k].entrylo1);

	}
	
}

void hpexec_clear_tlb(unsigned long virt, uint64_t phys, uint64_t size)
{
	uint64_t mapped_size, this_size;
	struct xlr_vcpu_wakeup_info *t;
	int mycpu = processor_id(), i;
	uint32_t cpu_map = cpu_buddy_map;
	extern void send_ipi(uint32_t mask, uint32_t ipi);

	mapped_size = 0;
	while (size > mapped_size) {
		this_size = find_tlb_alignment(phys + mapped_size, 
						    size - mapped_size);
		dbg_printf("hpexec_clear_tlb... phys %llx, virt %lx, size %llx\n",
		       (unsigned long long)(phys + mapped_size),  virt + mapped_size, 
		       (unsigned long long)this_size);
		clear_tlb_in_pte(virt + mapped_size, phys + mapped_size, this_size/2);
		mapped_size += this_size;
	}

	
	flushtlb_svaddr = virt;
	flushtlb_evaddr = virt + size;

	/* Flush on my cpu */
	flush_tlb_range(flushtlb_svaddr, flushtlb_evaddr, MIN_PAGE_SIZE + MIN_PAGE_SIZE);

	if(cpu_map != 0) {
		cpu_map &= (~(1 << mycpu));
		/* Send ipis to all other than me */
		for(i = 0; i < MAX_CPUS; i++) {
			if(!((1 << i) & cpu_map))
				continue;
			t = (struct xlr_vcpu_wakeup_info *) (xlr_vcpu_wakeup_info + i);
			if(t->cpu_status != CPU_RUNNING)
				continue;
			atomic_add(1, &flushtlb_cpucnt);
			dbg_printf("sending ipi to cpu %x\n", (1 << i));
			send_ipi((1<< i), IRQ_TLB_FLUSH_IPI);
		}
		while(atomic_read(&flushtlb_cpucnt) != 0);
	}
	dbg_printf("Tlb flush is done\n");

}

void tlb_flush_intr_handler(int irq)
{
	dbg_printf("%s in cpu %d\n", __FUNCTION__, processor_id());
	flush_tlb_range(flushtlb_svaddr, flushtlb_evaddr, MIN_PAGE_SIZE + MIN_PAGE_SIZE);
	atomic_sub(1, &flushtlb_cpucnt);
}

/* Dump the current entry* and pagemask registers */
static void dump_cur_tlb_regs(void)
{
	unsigned int entryhihi, entryhilo, entrylo0hi, entrylo0lo, entrylo1hi;
	unsigned int entrylo1lo, pagemask;

	__asm__ __volatile__(".set push             \n"
			".set noreorder        \n"
			".set mips64           \n"
			".set noat             \n"
			"     tlbr             \n"
			"     dmfc0  $1, $10   \n"
			"     dsrl32 %0, $1, 0 \n"
			"     sra    %1, $1, 0 \n"
			"     dmfc0  $1, $2    \n"
			"     dsrl32 %2, $1, 0 \n"
			"     sra    %3, $1, 0 \n"
			"     dmfc0  $1, $3    \n"
			"     dsrl32 %4, $1, 0 \n"
			"     sra    %5, $1, 0 \n"
			"     mfc0   %6, $5    \n"
			".set pop              \n":"=r"(entryhihi),
			"=r"(entryhilo),
			"=r"(entrylo0hi),
			"=r"(entrylo0lo),
			"=r"(entrylo1hi),
			"=r"(entrylo1lo), "=r"(pagemask));
	printf("%08X%08X %08X%08X %08X%08X %08X",
			entryhihi, entryhilo,
			entrylo0hi, entrylo0lo, entrylo1hi, entrylo1lo, pagemask);
}

void nlm_dump_tlb(void)
{
	int entry;
	int tlbsize;
	unsigned int config1;

	config1 = read_c0_config1();
	tlbsize = ((config1 >> 25) & 0x3f) + 1;

	printf("\n\nFull TLB Dump on cpu %d:\n"
		"Idx      EntryHi       EntryLo0          EntryLo1     PageMask\n"
		"--------------------------------------------------------------\n", processor_id());
	for (entry = 0; entry < tlbsize; entry++) {
		write_c0_index(entry);
		printf("\n%02i ", entry);
		__asm__ __volatile__(".set push             \n"
				".set mips64           \n"
				"     tlbr             \n"
				".set pop              \n");
		dump_cur_tlb_regs();
	}
	printf("\n");
}

